package com.ftdi.uart2gpio.ui;

import java.awt.Component;
import java.awt.GridLayout;
import java.awt.Font;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;

import javax.swing.JFrame;
import javax.swing.JScrollPane;
import javax.swing.JButton;
import javax.swing.ButtonGroup;
import javax.swing.Icon;
import javax.swing.ImageIcon;
import javax.swing.JCheckBox;
import javax.swing.JCheckBoxMenuItem;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.JPopupMenu;
import javax.swing.JRadioButtonMenuItem;
import javax.swing.JTextArea;
import javax.swing.text.BadLocationException;
import javax.swing.text.DefaultCaret;

import com.ftdi.FT90x;
import com.ftdi.FT90x.DIRECTION;
import com.ftdi.io.SerialPortListener;
import com.ftdi.swing.IconStore;
import com.ftdi.util.AppProperties;

import java.util.concurrent.Semaphore;

import javax.swing.UIManager;


public class MainUI {

	/** The number of GPIOs to render */
	private static final int NUM_GPIOS = 67;
	/** The number of GPIO rows to render */
	private static final int GPIO_ROWS = 9;
	/** The number of GPIO columns to render */
	private static final int GPIO_COLS = 8;
	/** The copyright notice shown at the bottom of the window */
	private static final String COPYRIGHT = AppProperties.getProperties().getProperty("ftdi.copyright");
	
	private final FT90x bridge;
	
	private boolean[] bGpioPullupsEnabled = new boolean[NUM_GPIOS];
	private boolean[] bGpioIsInput = new boolean[NUM_GPIOS];
	
	private static final int[] IGNORE_GPIOS = {FT90x.GPIO_UART0_TXD, FT90x.GPIO_UART0_RXD};
	
	private static final int[] UPDATE_SPEEDS_MS = {5, 10, 25, 100, 125, 250, 500, 1000, 2000, 5000};
	private int iRxUpdateSpeedms = 5000;
	private static final int UPDATE_SPEED_DEFAULT_IDX = 5;
	
	/** Set this to a positive value to have the monitor/console only keep x amount of lines */
	private static final int MAX_SCROLLBACK_LINES = -1;
	
	/* UI Elements */
	private final String title;
	private final String port;
	
	private JFrame frame;
	
	private JCheckBox[] chkbGpiosStatus = new JCheckBox[NUM_GPIOS];
	
	private JComboBox<String> comboUpdateSpeed = new JComboBox<String>();
	
	private final Icon iconFtdi = IconStore.LoadIcon("/gfx/ftdiico.png");
	private final Icon iconOutputOn = IconStore.LoadIcon("/gfx/on_green.png");
	private final Icon iconOutputOff = IconStore.LoadIcon("/gfx/off_green.png");
	private final Icon iconInputOn = IconStore.LoadIcon("/gfx/on_yellow.png");
	private final Icon iconInputOff = IconStore.LoadIcon("/gfx/off_yellow.png");
	
	/* Serial Port Related */
	private final Semaphore semSerial = new Semaphore(1, true);
	private final Thread threadSerialInputMonitor = new Thread(new SerialInputMonitor());
	
	private final JTextArea taSerial = new JTextArea();
	
	
	/**
	 * Create the application.
	 * @param title The title of the window
	 * @param port The name of the Serial Port to use
	 */
	public MainUI(String title, String port) {
		this.port = port;
		this.title = title;
		this.bridge = new FT90x(port);
				
		initialize();
		
		bridge_init(port);
		threadSerialInputMonitor.start();
		
		frame.addWindowListener(new UIWindowListener());
		frame.setVisible(true);
	}
	
	/**
	 * Initialize the contents of the frame.
	 */
	private void initialize() {
		frame = new JFrame();
		frame.setResizable(false);
		frame.setTitle(title);
		frame.setBounds(0, 0, 505, 358);
		frame.setLocationRelativeTo(null);
		frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		frame.getContentPane().setLayout(null);
		frame.setIconImage(((ImageIcon)iconFtdi).getImage());
		
		/* Set up the GPIO Grid */
		JPanel panelGpio = new JPanel();
		panelGpio.setAlignmentX(Component.LEFT_ALIGNMENT);
		GridLayout gl_panelGpio = new GridLayout(GPIO_ROWS, GPIO_COLS);
		gl_panelGpio.setVgap(10);
		gl_panelGpio.setHgap(10);
		panelGpio.setLayout(gl_panelGpio);
		
		for (int gpio = 0; gpio < NUM_GPIOS; ++gpio)
		{
			bGpioPullupsEnabled[gpio] = false;
			bGpioIsInput[gpio] = false;
			
			JPanel p = new JPanel();
			p.setLayout(null);
						
			/* State Checkbox */
			chkbGpiosStatus[gpio] = new JCheckBox(""+gpio);
			chkbGpiosStatus[gpio].setBounds(0,0,60,16);
			chkbGpiosStatus[gpio].setFont(chkbGpiosStatus[gpio].getFont().deriveFont(
					chkbGpiosStatus[gpio].getFont().getStyle() | Font.BOLD, 
					chkbGpiosStatus[gpio].getFont().getSize() + 2f
				));
			
			if (gpioIsIgnored(gpio))
			{
				chkbGpiosStatus[gpio].setIcon(iconOutputOff);
				chkbGpiosStatus[gpio].setSelectedIcon(iconOutputOff);
				chkbGpiosStatus[gpio].setEnabled(false);
			}
			else
			{
				chkbGpiosStatus[gpio].setIcon(iconOutputOff);
				chkbGpiosStatus[gpio].setSelectedIcon(iconOutputOn);
				chkbGpiosStatus[gpio].setDisabledIcon(iconInputOff);
				chkbGpiosStatus[gpio].setDisabledSelectedIcon(iconInputOn);
				
				chkbGpiosStatus[gpio].addActionListener(new GpioStatusCBActionListener(gpio));
				
				/* Add RMB Listener */	
				chkbGpiosStatus[gpio].addMouseListener(new GpioPopupClickListener(gpio));
				
				if (bGpioIsInput[gpio] == true)
				{
					chkbGpiosStatus[gpio].setEnabled(false);
				}
				
			}

			p.add(chkbGpiosStatus[gpio]);
			panelGpio.add(p);
		}

		panelGpio.setBounds(5, 5, 400, 300);
		frame.getContentPane().add(panelGpio);
		
		/* Serial Monitor */
		taSerial.setFont(UIManager.getFont("TextPane.font").deriveFont(
					UIManager.getFont("TextPane.font").getStyle(),
					UIManager.getFont("TextPane.font").getSize() + 2f
				));
		taSerial.setEditable(false);
		taSerial.setAutoscrolls(true);
		
		DefaultCaret caret = (DefaultCaret)taSerial.getCaret();
		caret.setUpdatePolicy(DefaultCaret.ALWAYS_UPDATE);
		
		JLabel lblSerialPort = new JLabel(port+":");
		lblSerialPort.setBounds(410,3,89,12);
		lblSerialPort.setFont(lblSerialPort.getFont().deriveFont(
				lblSerialPort.getFont().getStyle() | Font.BOLD
			));
		frame.getContentPane().add(lblSerialPort);
						
		JScrollPane sp = new JScrollPane(taSerial);
		sp.setBounds(410, 18, 89, 237);
		sp.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);
		frame.getContentPane().add(sp);
		
		JButton butSpClear = new JButton("Clear");
		butSpClear.setBounds(410, 255, 89, 25);
		butSpClear.addActionListener(new ActionListener(){
			@Override
			public void actionPerformed(ActionEvent e) {
				taSerial.setText("");
			}
		});
		frame.getContentPane().add(butSpClear);
		
		JButton butSpScript = new JButton("Script...");
		butSpScript.setBounds(410, 280, 89, 25);
		butSpScript.addActionListener(new ScriptButtonActionListener());
		frame.getContentPane().add(butSpScript);
				
		/* Status and Options */
		JLabel lblCopy = new JLabel(COPYRIGHT);
		lblCopy.setEnabled(false);
		lblCopy.setBounds(5, 310, 300, 18);
		frame.getContentPane().add(lblCopy);
		
		for (int i: UPDATE_SPEEDS_MS)
			comboUpdateSpeed.addItem(i+" ms");
		comboUpdateSpeed.setBounds(410, 305, 89, 25);
		comboUpdateSpeed.setSelectedIndex(UPDATE_SPEED_DEFAULT_IDX);
		iRxUpdateSpeedms = UPDATE_SPEEDS_MS[UPDATE_SPEED_DEFAULT_IDX];
		comboUpdateSpeed.addActionListener(new ComboUpdateSpeedActionListener());
		
		JLabel lblUpdateSpeed = new JLabel("Input Polling Rate");
		lblUpdateSpeed.setBounds(303, 310, 105, 18);
		lblUpdateSpeed.setHorizontalAlignment(JLabel.RIGHT);
		
		frame.getContentPane().add(lblUpdateSpeed);
		frame.getContentPane().add(comboUpdateSpeed);
	}
	
	/**
	 * See if the GPIO is ignored for whatever reason
	 */
	private boolean gpioIsIgnored(int gpio)
	{
		for(int i:IGNORE_GPIOS)
		{
			if (i==gpio) return true;
		}
		return false;
	}
	
	/**
	 * Open the Serial Port and initialise the Bridge
	 * @param port The name of the Serial Port to use
	 */
	private void bridge_init(String port)
	{
		/* Reset, Turn command echoing off */
		bridge.reset();
		bridge.echo(false); 
		bridge.flush();
		
		/* Setup all the ports based on their current UI state */
		for(int i=0; i<NUM_GPIOS; ++i)
		{
			if (!gpioIsIgnored(i))
			{
				if (bGpioIsInput[i])
				{
					bridge.dir(i, FT90x.DIRECTION.INPUT);
					bridge.flush();
				}
				else
				{
					bridge.dir(i, FT90x.DIRECTION.OUTPUT);
					bridge.set(i, false);
					bridge.flush();
				}

			}
		}
		bridge.clear();
		
		
		try {
			Thread.sleep(1);
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
		SerialPortListener spl = new SerialPortMonitorUpdater();
		bridge.addSerialReceiveListener(spl);
		bridge.addSerialTransmitListener(spl);
		
	}
	
	/** Window Listener
	 * 
	 *  This window listener is used to tear down the Serial Port
	 *  and threads on close
	 */
	class UIWindowListener implements WindowListener {
		@Override
		public void windowActivated(WindowEvent e) {}
		@Override
		public void windowClosed(WindowEvent e) {}
		@Override
		public void windowClosing(WindowEvent e) {
			//System.out.println(this.getClass().getName()+": windowClosing()");
			threadSerialInputMonitor.interrupt();
			try {
				threadSerialInputMonitor.join(1000);
			} catch (InterruptedException ex) {
				
			} finally {
				bridge.close();
			}
		}
		@Override
		public void windowDeactivated(WindowEvent e) {}
		@Override
		public void windowDeiconified(WindowEvent e) {}
		@Override
		public void windowIconified(WindowEvent e) {}
		@Override
		public void windowOpened(WindowEvent e) {}
	}
	
	
	/** 
	 * GPIO Panel Right Click Menu
	 */
	@SuppressWarnings("serial")
	class GpioPopupMenu extends JPopupMenu {

		//private JMenuItem itemGpioNum;
		private JMenuItem itemGpioDesc;
		
		private JRadioButtonMenuItem itemrbGpioInput;
		private JRadioButtonMenuItem itemrbGpioOutput;
		
		private JCheckBoxMenuItem itemcbGpioPull;
		
		public GpioPopupMenu(int gpio) {
			//itemGpioNum = new JMenuItem("GPIO"+gpio);
			//itemGpioNum.setEnabled(false);
			itemGpioDesc = new JMenuItem(FT90x.getPinFunction(gpio));
			itemGpioDesc.setEnabled(false);
			
			itemcbGpioPull = new JCheckBoxMenuItem("Pull-ups");
			itemcbGpioPull.addActionListener(new GpioPopupMenuPullupsActionCommand(gpio));
			itemcbGpioPull.setSelected(bGpioPullupsEnabled[gpio]);
			
			itemrbGpioInput = new JRadioButtonMenuItem("Input");
			itemrbGpioInput.addActionListener(new GpioPopupMenuIOActionCommand(gpio, true));
			itemrbGpioInput.setSelected(bGpioIsInput[gpio]);
					
			itemrbGpioOutput = new JRadioButtonMenuItem("Output");
			itemrbGpioOutput.addActionListener(new GpioPopupMenuIOActionCommand(gpio, false));
			itemrbGpioOutput.setSelected(!bGpioIsInput[gpio]);
			
			ButtonGroup bg1 = new ButtonGroup();
			bg1.add(itemrbGpioInput);
			bg1.add(itemrbGpioOutput);
			
			//add(itemGpioNum);
			add(itemGpioDesc);
			addSeparator();
			add(itemrbGpioInput);
			add(itemrbGpioOutput);
			addSeparator();
			add(itemcbGpioPull);
		}
	}
	
	
	/** 
	 * GpioPopupMenu Input/Output Action Listener 
	 */
	class GpioPopupMenuIOActionCommand implements ActionListener {
		private final int gpio;
		private final boolean val;
		
		public GpioPopupMenuIOActionCommand(int gpio, boolean isInput) {
			this.gpio = gpio;
			this.val = isInput;
		}
		
		@Override
		public void actionPerformed(ActionEvent e) {

			try {
				semSerial.acquire();
				if (val == true && !bGpioIsInput[gpio])
				{
					// Input
					bridge.dir(gpio, DIRECTION.INPUT);
				}
				else if (val == false && bGpioIsInput[gpio])
				{
					// Output
					bridge.dir(gpio, DIRECTION.OUTPUT);
					bridge.set(gpio, false);
					chkbGpiosStatus[gpio].setSelected(false);
				}
				semSerial.release();
			} catch (InterruptedException ex) {
				ex.printStackTrace();
			}
			
			chkbGpiosStatus[gpio].setEnabled(!val);
			
			bGpioIsInput[gpio] = val;
		}
	}
	
	
	/**
	 * GpioPopupMenu Pullups Action Listener
	 */
	class GpioPopupMenuPullupsActionCommand implements ActionListener {
		private final int gpio;
		
		public GpioPopupMenuPullupsActionCommand(int gpio) {
			this.gpio = gpio;
		}
		
		@Override
		public void actionPerformed(ActionEvent e) {
			JCheckBoxMenuItem src = (JCheckBoxMenuItem) e.getSource();
			try {
				semSerial.acquire();
				bridge.pullups(gpio, src.isSelected());
				semSerial.release();
			} catch (InterruptedException ex) {
				ex.printStackTrace();
			}
			bGpioPullupsEnabled[gpio] = src.isSelected();
		}
	}
	
	
	/** 
	 * GpioPopupMenu Mouse Adaptor 
	 */
	class GpioPopupClickListener extends MouseAdapter {
		private final int gpio;
		
		public GpioPopupClickListener(int gpio) {
			this.gpio = gpio;
		}
		
		@Override
		public void mousePressed(MouseEvent e) {
			super.mousePressed(e);
			
			if (e.isPopupTrigger())
				doPopup(e);
		}
		
		@Override
		public void mouseReleased(MouseEvent e) {
			super.mouseReleased(e);
			
			if (e.isPopupTrigger())
				doPopup(e);
		}
		
		private void doPopup(MouseEvent e) {
			GpioPopupMenu menu = new GpioPopupMenu(this.gpio);
			menu.show(e.getComponent(), e.getX(), e.getY());
		}
	}
	
	
	/**
	 * Gpio Status Checkbox Action Listener
	 */
	class GpioStatusCBActionListener implements ActionListener {
		private final int gpio;
		
		public GpioStatusCBActionListener(int gpio) {
			this.gpio = gpio;
		}
		
		@Override
		public void actionPerformed(ActionEvent e) {
			JCheckBox src = (JCheckBox) e.getSource();
			try {
				semSerial.acquire();
				bridge.set(gpio, src.isSelected());
				semSerial.release();
			} catch (InterruptedException ex) {
				ex.printStackTrace();
			}
		}
	}
	
	
	/**
	 * Update Speed Combo Box Action Listener
	 */
	class ComboUpdateSpeedActionListener implements ActionListener {
		public ComboUpdateSpeedActionListener() {
		}
		
		@Override
		public void actionPerformed(ActionEvent e) {
			@SuppressWarnings("unchecked")
			JComboBox<String> src = (JComboBox<String>) e.getSource();
			
			iRxUpdateSpeedms = UPDATE_SPEEDS_MS[src.getSelectedIndex()];
			System.out.println("Changing Input Update Rate to "+iRxUpdateSpeedms+" ms");
		}
	}
	
	/**
	 * Serial Port Transmit and Receive Listener
	 */
	class SerialPortMonitorUpdater implements SerialPortListener {
		@SuppressWarnings("unused")
		@Override
		public void byteTransfered(byte c) {
			taSerial.append(Character.toString((char)c));
			if (MAX_SCROLLBACK_LINES > 0) {
				if (taSerial.getLineCount() > MAX_SCROLLBACK_LINES) {
					try {
						int end = taSerial.getLineEndOffset(0);
						taSerial.replaceRange("", 0, end);
					} catch (BadLocationException ex) {
						
					}
				}
			}
		}
		@Override
		public void onBreak() {}
		@Override
		public void onCTS() {}
		@Override
		public void onDSR() {}
		@Override
		public void onRing() {}
	}	
	
	/**
	 * Serial Port Input Monitor
	 * 
	 * This thread manages the reading of values for pins set as inputs
	 */
	class SerialInputMonitor implements Runnable {
		
		//private static final int SLEEP_BETWEEN_PINS_MS = 0;
		private boolean keepgoing = true;
		
		public SerialInputMonitor() {
			
		}
		
		@Override
		public void run() {
			System.out.println("Serial Input Monitor Started");
			
			try {
				while (keepgoing)
				{
					for (int i = 0; i < NUM_GPIOS && keepgoing; ++i)
					{
						semSerial.acquire();
						if (bGpioIsInput[i])
						{
							boolean val = bridge.get(i);
							//System.out.println("Read "+i+" and got "+val);
							chkbGpiosStatus[i].setSelected(val);
							//Thread.sleep(SLEEP_BETWEEN_PINS_MS);
						}
						semSerial.release();
					}
					Thread.sleep(iRxUpdateSpeedms);
				}
			} catch (InterruptedException ex) {
				//System.out.println("Serial Input Monitor Interrupted");
				keepgoing = false;
			}
			System.out.println("Serial Input Monitor Exiting");
			
		}
	}
	
	/**
	 * Script parser
	 */
	class ScriptButtonActionListener implements ActionListener {
		@Override
		public void actionPerformed(ActionEvent e) {
			final ScriptUI script = new ScriptUI(frame);
			Thread t = new Thread(new Runnable(){
				@Override
				public void run() {
					String text = script.getText();
					
					if (text != null)
					{
						System.out.println("Script: Got back the text...\r\n"+text);
						try {							
							semSerial.acquire();

							new ScriptParser(text).run();

							semSerial.release();
						} catch (InterruptedException ex) {
							ex.printStackTrace();
						}
					}
				}
			});
			t.start();
		}
	}
	
	class ScriptParser implements Runnable {
		private final String text;
		
		public ScriptParser(String text) {
			this.text = text;
		}
		
		public void run() {
			for (String line:text.split(FT90x.LINEEND))
				processLine(line);
		}
		
		public void processLine(String s) {
			int pin = 0;
			System.out.println("Script: Parsing \""+s+"\"");

			if (!s.isEmpty())
			{
				// Parse
				switch (s.charAt(0))
				{
				case FT90x.CMD_HIGH:
				case FT90x.CMD_LOW:
				case FT90x.CMD_READ:
				case FT90x.CMD_INPUT:
				case FT90x.CMD_OUTPUT:
				case FT90x.CMD_PULLUP_ON:
				case FT90x.CMD_PULLUP_OFF:
					if (s.length() > 1)
					{
						pin = Integer.parseInt(s.substring(1).trim());
						System.out.println("Script: Parsed Integer \""+s.substring(1)+"\" to "+pin);
					}
					else
					{
						System.err.println("Script: Missing pin parameter");
						return;
					}
					break;
				case FT90x.CMD_ECHO_ON:
				case FT90x.CMD_ECHO_OFF:
				case FT90x.CMD_HELP:
				case FT90x.CMD_RESET:
					System.err.println("Script: Forbidden Command "+s.charAt(0));
					return;
				default:
					System.err.println("Script: Unknown Command "+s.charAt(0));
					return;
				}

				if (gpioIsIgnored(pin))
				{
					System.err.println("Script: Pin "+pin+" cannot be used");
					return;
				}

				// Execute
				switch (s.charAt(0))
				{
				case FT90x.CMD_HIGH:
					System.out.println("Script: Set "+pin+" high");

					chkbGpiosStatus[pin].setSelected(true);
					bridge.write(s+FT90x.LINEEND);
					break;
				case FT90x.CMD_LOW:
					System.out.println("Script: Set "+pin+" low");

					chkbGpiosStatus[pin].setSelected(false);
					bridge.write(s+FT90x.LINEEND);
					break;
				case FT90x.CMD_READ:
					break;
				case FT90x.CMD_INPUT:
					System.out.println("Script: Set "+pin+" input");

					chkbGpiosStatus[pin].setEnabled(false);
					bGpioIsInput[pin] = true;
					bridge.write(s+FT90x.LINEEND);
					break;
				case FT90x.CMD_OUTPUT:
					System.out.println("Script: Set "+pin+" output");

					chkbGpiosStatus[pin].setEnabled(true);
					bGpioIsInput[pin] = false;
					bridge.write(s+FT90x.LINEEND);
					break;
				case FT90x.CMD_PULLUP_ON:
					System.out.println("Script: Set "+pin+" pullups enabled");

					bGpioPullupsEnabled[pin] = true;
					bridge.write(s+FT90x.LINEEND);
					break;
				case FT90x.CMD_PULLUP_OFF:
					System.out.println("Script: Set "+pin+" pullups disabled");

					bGpioPullupsEnabled[pin] = false;
					bridge.write(s+FT90x.LINEEND);
					break;
				default:
					System.err.println("Shouldn't get here");
					break;
				}
			}
		}
	}
	
}

